import sys
from poodle import planned
from logzero import logger
from kalc.model.system.base import HasLimitsRequests, HasLabel
from kalc.model.system.Scheduler import Scheduler
from kalc.model.system.globals import GlobalVar
from kalc.model.system.primitives import TypeServ
from kalc.model.system.Controller import Controller
from kalc.model.system.primitives import Label
from kalc.model.kinds.Service import Service
from kalc.model.kinds.Deployment import Deployment
from kalc.model.kinds.DaemonSet import DaemonSet
from kalc.model.kinds.Pod import Pod
from kalc.model.kinds.Node import Node
from kalc.model.kinds.PriorityClass import PriorityClass, zeroPriorityClass
from kalc.misc.const import *
from kalc.misc.problem import ProblemTemplate
from kalc.misc.util import cpuConvertToAbstractProblem, memConvertToAbstractProblem
from kalc.misc.script_generator import move_pod_with_deployment_script_simple

class KubernetesModel(ProblemTemplate):
    def register_goal(self, lambda_fun):
        self.lambda_goal.append(lambda_fun)

    def add_goal_in(self, goal_entry):
        self.goals_in.extend(goal_entry)

    def add_goal_eq(self, goal_entry):
        self.goals_eq.extend(goal_entry)

    def generate_goal(self):
        self.add_goal_eq([[self.scheduler.status, STATUS_SCHED["Clean"]]])

    def goal(self):
        for l in self.lambda_goal:
            assert l()
        if self.goals_in:
            for what, where in self.goals_in:
                assert what in where
        if self.goals_eq:
            for what1, what2 in self.goals_eq:
                assert what1 == what2
       

    def problem(self):
        self.scheduler = next(filter(lambda x: isinstance(x, Scheduler), self.objectList))
        self.globalVar = next(filter(lambda x: isinstance(x, GlobalVar), self.objectList))

    @planned(cost=1)
    def Mark_service_as_started(self,
                service1: Service,
                scheduler: "Scheduler",
                globalVar: GlobalVar
         ):
        assert globalVar.block_policy_calculated == False
        assert service1.amountOfActivePods > 0
        assert service1.isNull == False
        # assert globalVar.block_node_outage_in_progress == False
        service1.status = STATUS_SERV["Started"]

    @planned(cost=1)
    def Fill_priority_class_object(self,
            pod: "Pod",
            pclass: PriorityClass,
            globalVar: GlobalVar
         ):
        assert globalVar.block_policy_calculated == False
        # assert globalVar.block_node_outage_in_progress == False
        assert pod.spec_priorityClassName == pclass.metadata_name
        pod.priorityClass = pclass

        
    @planned(cost=1)
    def SetDefaultMemRequestForPod(self,
        pod1: "Pod",
        memLimit: int,
        globalVar: GlobalVar
         ):
            assert globalVar.block_policy_calculated == False
            # assert globalVar.block_node_outage_in_progress == False
            assert pod1.memRequest == -1
            assert pod1.memLimit > -1
            assert memLimit == pod1.memLimit

            pod1.memRequest = memLimit

            
    @planned(cost=1)
    def SetDefaultCpuRequestForPod(self,
        pod1: "Pod",
        cpuLimit: int,
        globalVar: GlobalVar
         ):
            assert globalVar.block_policy_calculated == False
            # assert globalVar.block_node_outage_in_progress == False
            assert pod1.cpuLimit > -1
            assert pod1.cpuRequest == -1
            assert cpuLimit == pod1.cpuLimit

            pod1.cpuRequest = cpuLimit

           
    @planned(cost=1)
    def SetDefaultMemLimitForPod(self,
        pod1: "Pod",
        node: "Node" ,
        memCapacity: int,
        globalVar: GlobalVar
         ):
            assert globalVar.block_policy_calculated == False
            # assert globalVar.block_node_outage_in_progress == False
            assert pod1.memLimit == -1
            assert node == pod1.atNode
            assert memCapacity == node.memCapacity
            pod1.memLimit = memCapacity

            
    @planned(cost=1)
    def SetDefaultCpuLimitForPod(self,
        pod1: "Pod",
        node: "Node" ,
        cpuCapacity: int,
        globalVar: GlobalVar
         ):
            assert globalVar.block_policy_calculated == False
            # assert globalVar.block_node_outage_in_progress == False
            assert pod1.cpuLimit == -1
            assert node == pod1.atNode
            assert cpuCapacity == node.cpuCapacity

            pod1.cpuLimit = cpuCapacity

            
    @planned(cost=1)
    def SetDefaultMemLimitForPodBeforeNodeAssignment(self,
        pod1: "Pod",
        node: "Node" ,
        memCapacity: int,
        globalVar: GlobalVar
         ):
            assert globalVar.block_policy_calculated == False
            # assert globalVar.block_node_outage_in_progress == False
            assert pod1.memLimit == -1
            assert memCapacity == node.memCapacity
            pod1.toNode = node
            pod1.memLimit = memCapacity

            
    @planned(cost=1)
    def SetDefaultCpuLimitForPodBeforeNodeAssignment(self,
        pod1: "Pod",
        node: "Node" ,
        cpuCapacity: int,
        globalVar: GlobalVar
         ):
            assert globalVar.block_policy_calculated == False
            # assert globalVar.block_node_outage_in_progress == False
            assert pod1.cpuLimit == -1
            assert cpuCapacity == node.cpuCapacity
            pod1.toNode = node
            pod1.cpuLimit = cpuCapacity

    @planned(cost=1)
    def Evict_and_replace_less_prioritized_pod_byMEM(self,
                podPending: "Pod",
                podToBeReplaced: "Pod",
                nodeForPodPending: "Node" ,
                scheduler: "Scheduler",
                priorityClassOfPendingPod: PriorityClass,
                priorityClassOfPodToBeReplaced: PriorityClass,
                globalVar: GlobalVar
         ):
        assert globalVar.block_policy_calculated == False
        # assert globalVar.block_node_outage_in_progress == False
        assert podPending in scheduler.podQueue
        assert podPending.toNode == nodeForPodPending
        assert nodeForPodPending.isNull == False
        assert podToBeReplaced.atNode == nodeForPodPending
        assert podPending.status == STATUS_POD["Pending"]
        assert priorityClassOfPendingPod == podPending.priorityClass
        assert priorityClassOfPodToBeReplaced ==  podToBeReplaced.priorityClass
        # assert preemptionPolicyOfPendingPod == priorityClassOfPendingPod.preemptionPolicy
        # assert preemptionPolicyOfPodToBeReplaced == priorityClassOfPodToBeReplaced.preemptionPolicy
        # assert priorityClassOfPendingPod.preemptionPolicy == self.constSymbol["PreemptLowerPriority"]
        assert priorityClassOfPendingPod.priority > priorityClassOfPodToBeReplaced.priority
        assert podPending.memRequest > nodeForPodPending.memCapacity - nodeForPodPending.currentFormalMemConsumption
        assert podToBeReplaced.status == STATUS_POD["Running"]
        podToBeReplaced.status = STATUS_POD["Killing"]

    @planned(cost=1)
    def Evict_and_replace_less_prioritized_pod_byCPU(self,
        podPending: "Pod",
        podToBeReplaced: "Pod",
        nodeForPodPending: "Node" ,
        scheduler: "Scheduler",
        priorityClassOfPendingPod: PriorityClass,
        priorityClassOfPodToBeReplaced: PriorityClass,
        globalVar: GlobalVar
         ):
        assert globalVar.block_policy_calculated == False
        
        # assert globalVar.block_node_outage_in_progress == False
        assert podPending in scheduler.podQueue
        assert podPending.toNode == nodeForPodPending
        assert nodeForPodPending.isNull == False
        assert podToBeReplaced.atNode == nodeForPodPending
        assert podPending.status == STATUS_POD["Pending"]
        assert priorityClassOfPendingPod == podPending.priorityClass
        assert priorityClassOfPodToBeReplaced ==  podToBeReplaced.priorityClass
        # assert preemptionPolicyOfPendingPod == priorityClassOfPendingPod.preemptionPolicy
        # assert preemptionPolicyOfPodToBeReplaced == priorityClassOfPodToBeReplaced.preemptionPolicy
        # assert priorityClassOfPendingPod.preemptionPolicy == self.constSymbol["PreemptLowerPriority"]
        assert priorityClassOfPendingPod.priority > priorityClassOfPodToBeReplaced.priority
        assert podPending.cpuRequest > nodeForPodPending.cpuCapacity - nodeForPodPending.currentFormalCpuConsumption
        assert podToBeReplaced.status == STATUS_POD["Running"]
        podToBeReplaced.status = STATUS_POD["Killing"]

    @planned(cost=1)
    def Mark_Pod_As_Exceeding_Mem_Limits(self, podTobeKilled: "Pod",nodeOfPod: "Node",
        globalVar: GlobalVar
         ):
        assert globalVar.block_policy_calculated == False
        # assert globalVar.block_node_outage_in_progress == False
        assert podTobeKilled.memLimitsStatus == STATUS_LIM["Limit Met"]
        assert nodeOfPod == podTobeKilled.atNode
        assert podTobeKilled.memLimit <  podTobeKilled.currentRealMemConsumption
        nodeOfPod.AmountOfPodsOverwhelmingMemLimits += 1
        podTobeKilled.memLimitsStatus = STATUS_LIM["Limit Exceeded"]

        
    @planned(cost=1)
    def Mark_Pod_As_Not_Exceeding_Mem_Limits(self, podTobeReanimated: "Pod",
        nodeOfPod: "Node",
        globalVar: GlobalVar
         ):
        assert globalVar.block_policy_calculated == False
        # assert globalVar.block_node_outage_in_progress == False
        assert nodeOfPod == podTobeReanimated.atNode
        assert podTobeReanimated.memLimitsStatus == STATUS_LIM["Limit Exceeded"]
        assert nodeOfPod == podTobeReanimated.atNode
        assert podTobeReanimated.memLimit >  podTobeReanimated.currentRealMemConsumption
        nodeOfPod.AmountOfPodsOverwhelmingMemLimits -= 1
        podTobeReanimated.memLimitsStatus = STATUS_LIM["Limit Met"]

    @planned(cost=1)
    def MemoryErrorKillPodExceedingLimits(self,
        nodeOfPod: "Node" ,
        pod1TobeKilled: "Pod",
        globalVar: GlobalVar
         ):
        assert globalVar.block_policy_calculated == False
        # assert globalVar.block_node_outage_in_progress == False
        assert pod1TobeKilled.atNode == nodeOfPod
        assert nodeOfPod.memCapacity < nodeOfPod.currentRealMemConsumption
        assert pod1TobeKilled.memLimitsStatus == STATUS_LIM["Limit Exceeded"]
        pod1TobeKilled.status = STATUS_POD["Killing"]

        
    @planned(cost=1)
    def MemoryErrorKillPodNotExceedingLimits(self,
        nodeOfPod: "Node" ,
        podTobeKilled: "Pod",
        globalVar: GlobalVar
         ):
        assert globalVar.block_policy_calculated == False
        # assert globalVar.block_node_outage_in_progress == False
        assert podTobeKilled.atNode == nodeOfPod
        assert nodeOfPod.memCapacity < nodeOfPod.currentRealMemConsumption
        assert podTobeKilled.memLimitsStatus == STATUS_LIM["Limit Met"]
        podTobeKilled.status = STATUS_POD["Killing"]
    # @planned(cost=1)
    def KillPod(self,
        podBeingKilled : "Pod",
        nodeWithPod : "Node" ,
        serviceOfPod: "Service",
        scheduler: "Scheduler",
        pods_daemonset: DaemonSet,
        pods_deployment: Deployment,
        globalVar: GlobalVar
        ):
        if globalVar.block_node_outage_in_progress == False:
            nodeWithPod.currentFormalMemConsumption -= podBeingKilled.memRequest
            nodeWithPod.currentFormalCpuConsumption -= podBeingKilled.cpuRequest
        if podBeingKilled.hasService == True:
            assert podBeingKilled in serviceOfPod.podList
            serviceOfPod.amountOfActivePods -= 1
            serviceOfPod.amountOfPodsInQueue += 1
        if podBeingKilled.hasDeployment == True:
            assert podBeingKilled in pods_deployment.podList
            pods_deployment.amountOfActivePods -= 1
        if podBeingKilled.hasDaemonset == True:
            assert podBeingKilled in pods_daemonset.podList
            pods_daemonset.amountOfActivePods -= 1
        assert globalVar.block_policy_calculated == False
        assert podBeingKilled.atNode == nodeWithPod
        assert podBeingKilled.status == STATUS_POD["Running"]
        assert podBeingKilled.cpuRequest > -1 #TODO: check that number  should be moved to ariphmetics module from functional module
        assert podBeingKilled.memRequest > -1 #TODO: check that number  should be moved to ariphmetics module from functional module
        nodeWithPod.amountOfActivePods -= 1
        nodeWithPod.allocatedPodList.remove(podBeingKilled)
        nodeWithPod.allocatedPodList_length -= 1
        if podBeingKilled in nodeWithPod.directedPodList:
            nodeWithPod.directedPodList.remove(podBeingKilled)
            nodeWithPod.directedPodList_length -= 1
        podBeingKilled.status = STATUS_POD["Pending"]
        scheduler.podQueue.add(podBeingKilled)
        scheduler.status = STATUS_SCHED["Changed"] # commented, solves
        scheduler.queueLength += 1
        podBeingKilled.toNode = Node.NODE_NULL
        podBeingKilled.atNode = Node.NODE_NULL

        # scheduler.debug_var = True # TODO DELETEME
        #TODO: make sure that calculation excude situations that lead to negative number in the result
        ## assert podBeingKilled.amountOfActiveRequests == 0 #For Requests
        ## assert amountOfActivePodsPrev == serviceOfPod.amountOfActivePods

    @planned(cost=1)
    def AddNodeToSelector(self, 
        pod1: "Pod",
        selectedNode: "Node",
        globalVar: GlobalVar
         ):
        assert globalVar.block_policy_calculated == False
        # assert globalVar.block_node_outage_in_progress == False
        pod1.nodeSelectorList.add(selectedNode) 

    @planned(cost=1)
    def SelectNode(self, 
        pod1: "Pod",
        selectedNode: "Node",
        globalVar: GlobalVar
         ):
        assert globalVar.block_policy_calculated == False
        assert globalVar.block_node_outage_in_progress == False
        assert pod1.toNode == Node.NODE_NULL
        assert selectedNode in pod1.nodeSelectorList
        pod1.toNode = selectedNode    

    @planned(cost=1) # TODO
    def StartPod(self, 
            podStarted: "Pod",
            node: "Node",
            scheduler: "Scheduler",
            serviceTargetForPod: "mservice.Service",
            pods_deployment: Deployment,
            pods_daemonset: DaemonSet,
            globalVar: GlobalVar
        ):
        if globalVar.nodeSelectorsEnabled == True:
            assert node in podStarted.nodeSelectorList
        if podStarted.hasService == True:
            assert podStarted in serviceTargetForPod.podList
            serviceTargetForPod.amountOfPodsInQueue -= 1
            serviceTargetForPod.amountOfActivePods += 1
            serviceTargetForPod.status = STATUS_SERV["Started"]
        if podStarted.hasDeployment == True:
            assert podStarted in pods_deployment.podList
            pods_deployment.amountOfActivePods += 1
        if podStarted.hasDaemonset == True:
            assert podStarted in pods_daemonset.podList
            pods_daemonset.amountOfActivePods += 1
        # if podStarted.memRequest == -1 and podStarted.memLimit > -1:
        #     podStarted.memRequest = podStarted.memLimit
        # if podStarted.cpuRequest == -1 and podStarted.cpuLimit > -1:
        #     podStarted.cpuRequest = podStarted.cpuLimit

        # if podStarted.memRequest > -1 and podStarted.memLimit == -1:
        #     podStarted.memLimit = node.memCapacity
        # if podStarted.cpuRequest > -1 and podStarted.cpuLimit == -1:
        #     podStarted.cpuLimit = node.cpuCapacity

        # if podStarted.memRequest == -1 and podStarted.memLimit == -1:
        #     podStarted.memLimit = node.memCapacity
        #     podStarted.memRequest = 0
        # if podStarted.cpuRequest == -1 and podStarted.cpuLimit == -1:
        #     podStarted.cpuLimit = node.cpuCapacity
        #     podStarted.cpuRequest = 0
        #todo: Soft conditions are not supported yet ( prioritization of nodes :  for example healthy  nodes are selected  rather then non healthy if pod  requests such behavior 
            # assert globalVar.block_node_outage_in_progress == False
        assert globalVar.block_policy_calculated == False
        assert podStarted in scheduler.podQueue
        if podStarted.toNode != Node.NODE_NULL:
            assert node == podStarted.toNode
        assert node.status == STATUS_NODE["Active"]
        assert podStarted.cpuRequest > -1
        assert podStarted.memRequest > -1
        assert node.currentFormalCpuConsumption + podStarted.cpuRequest <= node.cpuCapacity
        assert node.currentFormalMemConsumption + podStarted.memRequest <= node.memCapacity
        node.currentFormalCpuConsumption += podStarted.cpuRequest
        node.currentFormalMemConsumption += podStarted.memRequest
        podStarted.atNode = node       
        # node.pods.add(podStarted) # for new method
        # for pod in node.pods: # -> (forall (?pod-1 - Pod) ... (and (Node-pods ?node ?pod-1)
        #     assert not podStarted in pod.anti_affinite_with #      (not (Pod-anti_affninite_with ?pod-1 ?podStarted))))
        scheduler.queueLength -= 1
        scheduler.podQueue.remove(podStarted)
        node.amountOfActivePods += 1
        node.allocatedPodList.add(podStarted)
        node.allocatedPodList_length += 1
        podStarted.status = STATUS_POD["Running"]      

    @planned(cost=1)
    def MoveRunningPodToAnotherNode(self,
        pod: "Pod",
        nodeFrom: "Node",
        nodeTo: "Node",
        scheduler: "Scheduler",
        globalVar: GlobalVar,
        deployment: Deployment):
        assert pod in deployment.podList # Only pods with deployments can be moved by kalc 
        assert deployment.amountOfActivePods > 1
        # assert globalVar.block_policy_calculated == False
        assert pod.atNode == nodeFrom
        # assert pod.status == STATUS_POD["Running"]
        assert pod.cpuRequest > -1 #TODO: check that number  should be moved to ariphmetics module from functional module
        assert pod.memRequest > -1 #TODO: check that number  should be moved to ariphmetics module from functional module
        assert nodeTo in pod.nodeSelectorList
        assert nodeTo.status == STATUS_NODE["Active"]
        assert nodeTo.currentFormalCpuConsumption + pod.cpuRequest <= nodeTo.cpuCapacity
        assert nodeTo.currentFormalMemConsumption + pod.memRequest <= nodeTo.memCapacity
        assert pod.antiaffinity_met == False
        assert pod.calc_antiaffinity_pods_list_length == 0

        nodeFrom.currentFormalMemConsumption -= pod.memRequest
        nodeFrom.currentFormalCpuConsumption -= pod.cpuRequest
        nodeFrom.amountOfActivePods -= 1
        nodeFrom.allocatedPodList.remove(pod)
        nodeFrom.allocatedPodList_length -= 1
        nodeTo.currentFormalCpuConsumption += pod.cpuRequest
        nodeTo.currentFormalMemConsumption += pod.memRequest
        pod.atNode = nodeTo
        nodeTo.amountOfActivePods += 1
        nodeTo.allocatedPodList.add(pod)
        nodeTo.allocatedPodList_length += 1

        self.script.append(move_pod_with_deployment_script_simple(pod, nodeTo, self.objectList))

        # if podStarted.memRequest == -1 and podStarted.memLimit > -1:
        #     podStarted.memRequest = podStarted.memLimit
        # if podStarted.cpuRequest == -1 and podStarted.cpuLimit > -1:
        #     podStarted.cpuRequest = podStarted.cpuLimit

        # if podStarted.memRequest > -1 and podStarted.memLimit == -1:
        #     podStarted.memLimit = node.memCapacity
        # if podStarted.cpuRequest > -1 and podStarted.cpuLimit == -1:
        #     podStarted.cpuLimit = node.cpuCapacity

        # if podStarted.memRequest == -1 and podStarted.memLimit == -1:
        #     podStarted.memLimit = node.memCapacity
        #     podStarted.memRequest = 0
        # if podStarted.cpuRequest == -1 and podStarted.cpuLimit == -1:
        #     podStarted.cpuLimit = node.cpuCapacity
        #     podStarted.cpuRequest = 0
        #todo: Soft conditions are not supported yet ( prioritization of nodes :  for example healthy  nodes are selected  rather then non healthy if pod  requests such behavior 
            # assert globalVar.block_node_outage_in_progress == False



    @planned(cost=1)
    def SchedulerCleaned(self, 
        scheduler: "Scheduler",
        globalVar: GlobalVar):
        assert globalVar.block_node_outage_in_progress == False
        assert  scheduler.queueLength == 0
        assert globalVar.block_policy_calculated == True
        scheduler.status = STATUS_SCHED["Clean"]

    @planned(cost=1)
    def Initiate_node_outage_searched(self,
        node_with_outage: "Node",
        globalVar: GlobalVar):
        assert globalVar.block_node_outage_in_progress == False
        # assert globalVar.amountOfNodesDisrupted < globalVar.limitOfAmountOfNodesDisrupted
        assert globalVar.is_node_disrupted == False # TODO: checking ONE node outage! Need fixing
        assert node_with_outage.searchable == True
        assert node_with_outage.status == STATUS_NODE["Active"]
        assert node_with_outage.isSearched == True
        node_with_outage.status = STATUS_NODE["Killing"]
        globalVar.block_node_outage_in_progress = True
        
    @planned(cost=1)
    def Initiate_killing_of_Pod_because_of_node_outage(self,
        node_with_outage: "Node",
        pod_killed: "podkind.Pod",
        globalVar: GlobalVar):
        assert globalVar.block_node_outage_in_progress == True
        assert pod_killed.status == STATUS_POD["Running"]
        assert pod_killed.atNode == node_with_outage
        assert node_with_outage.status == STATUS_NODE["Killing"]
        pod_killed.status = STATUS_POD["Killing"]
        
    @planned(cost=1)
    def NodeOutageFinished(self,
        node: "Node",
        globalVar: GlobalVar):
        assert globalVar.block_node_outage_in_progress == True
        assert node.amountOfActivePods == 0
        assert node.status == STATUS_NODE["Killing"]
        globalVar.amountOfNodesDisrupted += 1
        node.status = STATUS_NODE["Inactive"]
        globalVar.block_node_outage_in_progress = False
        globalVar.is_node_disrupted = True # TODO: checking ONE node outage! Need fixing
        # TODO make ability to calculate multiple nodes outage
        
    @planned(cost=1)
    def ReplaceNullCpuRequestsWithZero(self,
        pod: "Pod"):
        # assert pod.status == STATUS_POD["Running"]
        assert pod.cpuRequest == -1
        pod.cpuRequest = 0
        
    @planned(cost=1)
    def ReplaceNullMemRequestsWithZero(self,
        pod: "Pod"):
        # assert pod.status == STATUS_POD["Running"]
        assert pod.memRequest == -1
        pod.memRequest = 0




from kalc.model.full import kinds_collection
from collections import defaultdict
from kalc.model.search import KubernetesModel

class PolicyImplementer:
    def __init__(self, obj, all_policies, state_objects):
        self._target_object = obj
        self._all_policies = all_policies
        self._state_objects = state_objects
        self._instantiated_policies = {}
        self._sealed = False
        for p_name, p_class in all_policies.items():
            # print("Adding policy")
            p_obj = p_class(obj, state_objects)
            self._instantiated_policies[p_name] = p_obj
            setattr(self, p_name, p_obj)
        self._sealed = True
    
    def __setattr__(self, name, val):
        if name.startswith("_") or not self._sealed: return super().__setattr__(name, val)
        if name in self._instantiated_policies and self._instantiated_policies[name].TYPE == "property":
            self._instantiated_policies[name]._set(val)
        return super().__setattr__(name, val)
    
    def __getattr__(self, name):
        if name.startswith("_") or not self._sealed: return super().__getattr__(name)
        if name in self._instantiated_policies and self._instantiated_policies[name].TYPE == "property":
            self._instantiated_policies[name]._get()
        return super().__getattr__(name)
    
    def apply(self, kube: KubernetesModel):
        # TODO HERE: hypothesis must be added in some fashion to allow running combinations:
        # - add hypotheses to a list with its goals and order
        # - after we collected all hypoteses, take first N and generate basic run plan
        print("AAA trying to activate policy", self._instantiated_policies)
        hypotheses = []
        for pname, pobject in self._instantiated_policies.items():
            print("AAA checking ", pname)
            if pobject.activated:
                print("AAA found activated policy")
                for hname, hval in pobject.hypotheses.items():
                    print("AAA Adding hypothesis", hname)
                    hypotheses.append(hval)
                for name in dir(pobject):
                    if callable(getattr(pobject, name)) and hasattr(getattr(pobject, name), "_planned"):
                        print("AAA Adding planned method from policy", name)
                        kube.add_external_method(getattr(pobject, name))
        return hypotheses


class PolicyEngineClass:
    def __init__(self):
        self.registered_engines = defaultdict(dict)
        self.state_objects = []
        self._cache = {}

    def register_state_objects(self, state_objects):
        self.state_objects = state_objects

    def register(self, kind_name, policy_name, policy_class):
        self.registered_engines[kind_name.__name__][policy_name] = policy_class
        # print("Registered policy", self.registered_engines)
    
    def get(self, obj):
        if not obj in self._cache:
            self._cache[obj] = PolicyImplementer(obj, self.registered_engines[obj.__class__.__name__], self.state_objects)
        return self._cache[obj]

policy_engine = PolicyEngineClass()

for kind_name, kind_class in kinds_collection.items():
    kind_class.policy = "STUB"
    kind_class.policy_engine = policy_engine
    # if hasattr(kind_class, "__getattr__"):
    #     old_getattr = kind_class.__getattr__
    #     _getattr = True
    # else:
    #     old_getattr = kind_class.__getattribute__
    #     _getattr = False
    # def _get_policy_getattr(self, name):
    #     if name == "policy":
    #         return policy_engine.get(self)
    #     return old_getattr(self, name)
    # if _getattr: kind_class.__getattr__ = _get_policy_getattr
    # else: kind_class.__getattribute__ = _get_policy_getattr

# WARNING this whole thing looks too cryptic and must be re-implemented

class BasePolicy:
    TYPE = "function"
    ACTIVATED = False
    def __init__(self, obj, state_objects):
        self.target_object = obj
        self.hypotheses = {}
        self.state_objects = state_objects

    def register_hypothesis(self, name, func, order=1):
        self.hypotheses[name] = [order, name, func]
    
    def _set(self, val):
        if not self.ACTIVATED:
            self.ACTIVATED = True
            self.register()
        return self.set(val)
    
    def _get(self):
        return self.get()
    
    def set(self, val):
        raise NotImplementedError("Author has not implemented set() for this policy")

    def get(self):
        raise NotImplementedError("Author has not implemented get() for this policy")

#########################################################################
# Now import all policies from policies folder

import kalc.policies
import pkgutil, inspect

policies_collection = {}

for importer, modname, ispkg in pkgutil.iter_modules(
        path=kalc.policies.__path__,
        prefix=kalc.policies.__name__+'.'):
    # print("Importing", modname)
    module = __import__(modname, fromlist="dummy") # importing should be enough
    # for n in dir(module):
    #     c = getattr(module, n)
    #     if inspect.isclass(c) and issubclass(c, Object) and not c is Object:
    #         kinds_collection[c.__name__] = c


    


import pandas as pd

def label_string_to_python(label: str):
    return label.replace("_", "__").replace("-", "_")

def return_one_or_dataframe(matched: list):
    if len(matched) > 1: return KindFilterResult(matched)
    elif len(matched) == 1: return matched[0]
    else: return None

class KindPlaceholder:
    pass

class FilteredKind:
    def __init__(self, kind_name, state_objects):
        self._kind_name = kind_name
        self._state_objects = state_objects
        self._cache = []
    
    def _gen_listing(self):
        raise NotImplementedError()

    def _safe_getattr(self, value):
        return super().__getattribute__(value)

    def __dir__(self):
        # if not self.cache: self.cache = self.gen_listing()
        # return self.cache
        return [str(x) for x in self._gen_listing()]
    
    def __getattr__(self, value):
        if value.startswith("_"): return super().__getattribute__(value)
        return self._safe_getattr(value)
    
class KindFilterResultSeries(pd.Series):

    @property
    def _constructor(self):
        return KindFilterResultSeries 

    @property
    def _constructor_expanddim(self):
        return KindFilterResult 


class KindFilterResult(pd.DataFrame):

    @property
    def _constructor(self):
        return KindFilterResult 

    @property
    def _constructor_sliced(self):
        return KindFilterResultSeries 

class FilterByLabelValue(FilteredKind):
    def __init__(self, kind_name, state_objects, target_key):
        super().__init__(kind_name, state_objects)
        self._target_key = target_key

    def _gen_listing(self):
        label_values = set()
        for obj in filter(lambda x: x.__class__.__name__ == self._kind_name, self._state_objects):
            for label in obj.metadata_labels:
                if label.key == self._target_key:
                    label_values.add(label_string_to_python(label.value))
        return list(label_values)
 
    def _safe_getattr(self, val):
        # if not self._cache: self._cache = self.gen_listing()
        self._cache = self._gen_listing()
        if val in self._cache:
            matched = []
            for ob in filter(lambda x: x.__class__.__name__ == self._kind_name, self._state_objects):
                for l in ob.metadata_labels:
                    if label_string_to_python(l.key) == self._target_key and \
                                    label_string_to_python(l.value) == val:
                        matched.append(ob)
            return return_one_or_dataframe(matched)
        return super().__getattribute__(val)

class FilterByLabelKey(FilteredKind):
    def _gen_listing(self):
        label_keys = set()
        for obj in filter(lambda x: x.__class__.__name__ == self._kind_name, self._state_objects):
            for label in obj.metadata_labels:
                label_keys.add(label_string_to_python(label.key))
                # because of ipython __dir__ completion bug,
                # we will be setting actual attrs here
                setattr(self, label_string_to_python(label.key), \
                    FilterByLabelValue(self._kind_name, self._state_objects, \
                            target_key=label_string_to_python(label.key)))
        return list(label_keys)
    
    def _safe_getattr(self, val):
        self._gen_listing()
        return super().__getattribute__(val)

    # because of bug in __dir__ ipython support, this does not work. TODO: file bug 
    # def _safe_getattr(self, val):
    #     if not self._cache: self._cache = self._gen_listing()
    #     # self._cache = self._gen_listing()
    #     if val in self._cache:
    #         return FilterByLabelValue(self._kind_name, self._state_objects, target_key=val)
    #     return super().__getattribute__(val)


class FilterByName(FilteredKind):
    def _gen_listing(self, return_object=None):
        names = []
        for obj in filter(lambda x: x.__class__.__name__ == self._kind_name, self._state_objects):
            names.append(label_string_to_python(str(obj.metadata_name)))
            if return_object == label_string_to_python(str(obj.metadata_name)): 
                return obj
        return names
    
    def _safe_getattr(self, val):
        if val in self._gen_listing():
            return self._gen_listing(return_object=val)
        super().__getattribute__(val)




class ClientConfig(object):
    PUBLIC_KEY = 'VllVpxXLMCpE5Uuw1nqkLDPjGE85Pnmpdt8VIcmOr40'
    APP_NAME = 'kubectl-val'
    COMPANY_NAME = 'CriticalHop'
    HTTP_TIMEOUT = 6000
    MAX_DOWNLOAD_RETRIES = 3
    UPDATE_URLS = ['https://download.criticalhop.com/']



from typing import Set
from kalc.model.kinds.Pod import Pod
from kalc.model.kinds.Node import Node
from kalc.model.kinds.Service import Service
from kalc.model.system.Scheduler import Scheduler
from kalc.model.system.globals import GlobalVar
from kalc.misc.const import *
from poodle import planned
from kalc.policy import policy_engine, BasePolicy
from kalc.model.search import KubernetesModel

class StubPolicy(BasePolicy):
    TYPE = "property"

    def register(self):
        GlobalVar.register_property(name="stub", type=bool, default=False)

    def set(self, val: bool):
        assert isinstance(val, bool), "Can only accept True or False"

        if val:
            # enable

            def hypothesis_1(kube: KubernetesModel):
                # TODO: hypotheses can not work in parallel this way: will modify main object
                self.target_object.stub = False
                kube.register_goal(lambda: self.target_object.stub == True)
            
            self.register_hypothesis("Stub worked", hypothesis_1, order=1)
        else:
            # disable
            pass

    @planned(cost=1)
    def do_nothing(self,
            service: Service,
            globalVar: GlobalVar):
        assert globalVar.stub == False
        globalVar.stub = True

policy_engine.register(GlobalVar, "stub", StubPolicy)





import yaml, copy
import os, click
from logzero import logger
from collections import defaultdict
from kalc.misc.object_factory import labelFactory
from poodle import planned, Property, Relation
from kalc.misc.util import objwalk, find_property, k8s_to_domain_object, POODLE_MAXLIN, getint, split_yamldumps
from kalc.misc.util import cpuConvertToNorm, memConvertToNorm
from kalc.model.full import kinds_collection
from kalc.model.search import K8ServiceInterruptSearch
from kalc.model.system.globals import GlobalVar
from kalc.model.system.Scheduler import Scheduler
from kalc.model.kinds.ReplicaSet import ReplicaSet
import kalc.misc.util

KINDS_LOAD_ORDER = ["PriorityClass", "Service", "Node", "Pod", "ReplicaSet"]

class KubernetesCluster:
    CREATE_MODE = "create"
    LOAD_MODE = "load"
    SCALE_MODE = "scale"
    APPLY_MODE = "apply"
    REPLACE_MODE = "replace"
    REMOVE_MODE = "remove"

    def __init__(self):
        self.dict_states = defaultdict(list)
        self._reset()

    def _reset(self):
        "Reset object states and require a rebuild with _bulid_state"
        self.scheduler = Scheduler("Sheduler")
        self.globalvar = GlobalVar("GlobalVar")
        self.state_objects = [self.scheduler, self.globalvar]

    def load_dir(self, dir_path):
        for root, dirs, files in os.walk(dir_path):
            for fn in files:
                for ys in split_yamldumps(open(os.path.join(root, fn)).read()):
                    self.load(ys, self.LOAD_MODE)

    # load - load from dump , scale/apply/replace/remove/create - are modes from kubernetes
    def load(self, str_, mode=LOAD_MODE):
        for doc in yaml.load_all(str_, Loader=yaml.FullLoader):
            if len(str(doc)) < 100:
                click.echo(f"# WARNING  doc - {doc} to short!!!!")
            if "items" in doc:
                for item in doc["items"]: self.load_item(item, mode)
            else: self.load_item(doc, mode)


    def scale(self, replicas, str_):
        print("scale {0}".format(str_))
        # type = str_.split(".")[0] # e.g. Deployment 
        # for k,v in dict_states.items():
        #     for item in v:

    def load_item(self, item, mode=LOAD_MODE):
        assert isinstance(item, dict), item
        create = False
        if mode == self.CREATE_MODE : create = True
        item["__created"] = create
        item["__mode"] = mode
        item["__scale_replicas"] = 5
        self.dict_states[item["kind"]].append(item)

    def _build_item(self, item):
        try:
            obj = kinds_collection[item["kind"]]()
        except KeyError:
            logger.warning("Skipping unsupported kind %s" % item["kind"])
            return

        create = item["__created"]
        mode = item["__mode"]
        replicas = item["__scale_replicas"]
        obj.kubeguard_created = create # special property to distinguish "created"
        for prop in objwalk(item):
            p, val = find_property(obj, prop)
            if p is None: continue
            val = k8s_to_domain_object(val)
            if isinstance(getattr(obj, p), Relation):
                getattr(obj, p).add(val)
            elif isinstance(getattr(obj, p), Property):
                setattr(obj, p, val)
            else:
                # means has setter
                setattr(obj, p, val)
        obj.yaml = copy.deepcopy(item)
        #cleanup service information
        for y in item:
            if y[0:2] == '__':
                del(obj.yaml[y])
        obj.yaml_orig = copy.deepcopy(obj.yaml)
        if mode == self.CREATE_MODE and hasattr(obj, "hook_after_create"):
            obj.hook_after_create(self.state_objects)
        if mode == self.LOAD_MODE and hasattr(obj, "hook_after_load"):
            obj.hook_after_load(self.state_objects)
        if mode == self.APPLY_MODE and hasattr(obj, "hook_after_apply"):
            obj.hook_after_apply(self.state_objects)
        # if mode == self.SCALE_MODE and hasattr(obj, "hook_scale"):
        #     obj.hook_scale(self.state_objects, replicas)
        self.state_objects.append(obj)

    def _normalize_mappings(self):
        max_ram = 0
        max_cpu = 0
        priorities = []
        for k,v in self.dict_states.items():
            for item in v:
                if k == "Pod":
                    for cnt in item["spec"]["containers"]:
                        try:
                            ram = memConvertToNorm(cnt["resources"]["requests"]["memory"])
                            if max_ram < ram: max_ram = ram
                        except KeyError:
                            pass
                        try:
                            cpu = cpuConvertToNorm(cnt["resources"]["requests"]["cpu"])
                            if max_cpu < cpu: max_cpu = cpu
                        except KeyError:
                            pass
                    # TODO: check limits too
                if k == "Node":
                    try:
                        ram = memConvertToNorm(item["status"]["allocatable"]["memory"])
                        if max_ram < ram: max_ram = ram
                        cpu = cpuConvertToNorm(item["status"]["allocatable"]["cpu"])
                        if max_cpu < cpu: max_cpu = cpu
                    except KeyError:
                        pass
                if k == "PriorityClass":
                    priorities.append(int(item["value"]))
        if max_cpu > 0:
            kalc.misc.util.CPU_DIVISOR = int(max_cpu / (kalc.misc.util.POODLE_MAXLIN / 2))
        if max_ram > 0:
            kalc.misc.util.MEM_DIVISOR = int(max_ram / (kalc.misc.util.POODLE_MAXLIN / 2))
        if len(priorities):
            pri_map = {}
            i = 1
            for p in sorted(priorities):
                pri_map[p] = i
                i+=1
            kalc.misc.util.PRIO_MAPPING = pri_map
            assert i <= kalc.misc.util.POODLE_MAXLIN, "Too many different priorities"

    def _build_state(self):
        self._normalize_mappings()
        collected = self.dict_states.copy()
        for k in KINDS_LOAD_ORDER:
            if not k in collected: continue
            for item in collected[k]:
                self._build_item(item)
            del collected[k]
        for k,v in collected.items():
            for item in v:
                self._build_item(item)
        self._check()

    def _check(self):
        "Run internal checks"
        assert getint(self.scheduler.queueLength) < POODLE_MAXLIN, "Queue length overflow {0} < {1}".format(getint(self.scheduler.queueLength), POODLE_MAXLIN)

    def create_resource(self, res: str):
        self.load(res, mode=self.CREATE_MODE)

    def apply_resource(self, res: str):
        self.load(res, mode=self.APPLY_MODE)

    def fetch_state_default(self):
        "Fetch state from cluster using default method"
        raise NotImplementedError()

    def run(self):
        if len(self.state_objects) < 3: self._build_state()
        k = K8ServiceInterruptSearch(self.state_objects)
        k.run()
        self.plan = k.plan
        return self.plan
        # TODO: represent plan




import sys
import random
from kalc.model.system.base import ModularKind
from typing import Set
from kalc.model.system.primitives import Label, StatusNode
from kalc.model.system.base import HasLabel
from kalc.misc.util import cpuConvertToAbstractProblem, memConvertToAbstractProblem
from kalc.misc.const import STATUS_NODE
from kalc.model.system.globals import GlobalVar


class Node(ModularKind, HasLabel):
    # k8s attributes
    metadata_ownerReferences__name: str
    metadata_name: str
    spec_priorityClassName: str
    labels: Set[Label]
    # pods: Set[mpod.Pod]
    cpuCapacity: int
    memCapacity: int
    currentFormalCpuConsumption: int
    currentFormalMemConsumption: int
    currentRealMemConsumption: int
    currentRealCpuConsumption: int
    AmountOfPodsOverwhelmingMemLimits: int
    isNull: bool
    status: StatusNode
    amountOfActivePods: int
    searchable: bool
    isSearched: bool
    different_than: Set["Node"]
    allocatedPodList: Set["Pod"]
    allocatedPodList_length: int
    directedPodList: Set["Pod"]
    directedPodList_length: int

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.metadata_name = "modelNode"+str(random.randint(100000000, 999999999))
        # self.metadata_name = "model-default-name"
        self.AmountOfPodsOverwhelmingMemLimits = 0
        self.currentFormalCpuConsumption = 0
        self.currentFormalMemConsumption = 0
        self.currentRealCpuConsumption = 0
        self.currentRealMemConsumption = 0
        self.cpuCapacity = 0
        self.memCapacity = 0
        self.isNull = False
        self.status = STATUS_NODE["Active"]
        self.amountOfActivePods = 0
        self.searchable = True
        self.isSearched = False
        self.allocatedPodList_length = 0
        self.directedPodList_length = 0
        
    def hook_after_create(self, object_space):
        globalVar = next(filter(lambda x: isinstance(x, GlobalVar), object_space))
        globalVar.amountOfNodes += 1

        nodes = filter(lambda x: isinstance(x, Node), object_space)
        for node in nodes:
            if node != self:
                self.different_than.add(node)
                node.different_than.add(self)

    def hook_after_load(self, object_space):
        globalVar = next(filter(lambda x: isinstance(x, GlobalVar), object_space))
        globalVar.amountOfNodes += 1
        nodes = filter(lambda x: isinstance(x, Node), object_space)
        for node in nodes:
            if node != self:
                self.different_than.add(node)
                node.different_than.add(self)

    @property
    def status_allocatable_memory(self):
        pass
    @status_allocatable_memory.setter
    def status_allocatable_memory(self, value):
        self.memCapacity = memConvertToAbstractProblem(value)

    @property
    def status_allocatable_cpu(self):
        pass
    @status_allocatable_cpu.setter
    def status_allocatable_cpu(self, value):
        self.cpuCapacity = cpuConvertToAbstractProblem(value)

    def __str__(self):
        if str(self.metadata_name) == "None":
            return "<unnamed node>"
        return str(self.metadata_name)
    # def __repr__(self):
    #     return 'Nodename : ' + str(self._get_value()) 

Node.NODE_NULL = Node("NULL")
Node.NODE_NULL.isNull = True
Node.NODE_NULL.status = STATUS_NODE["Inactive"]
Node.NODE_NULL.metadata_name = "Null-Node"
Node.NODE_NULL.searchable = False




from kalc.model.system.primitives import Label

class _LabelFactory:
    def __init__(self):
        self.labels = {}
    def get(self, name, value):
        lbl = f"{name}:{value}"
        if not lbl in self.labels:
            self.labels[lbl] = Label(lbl)
        return self.labels[lbl]

labelFactory = _LabelFactory()



from poodle import Object
from kalc.model.system.primitives import Type, Status
from kalc.model.system.base import ModularKind
import kalc.model.kinds.Deployment as mdeployment
import kalc.model.kinds.Node as mnode
from typing import Set

class GlobalVar(ModularKind):
    is_deployment_disrupted: bool
    is_deployment_distuption_searchable: bool
    is_service_disrupted: bool
    is_service_disruption_searchable: bool
    is_node_disrupted: bool
    amountOfNodesDisrupted: int
    limitOfAmountOfNodesDisrupted: int
    amountOfNodes: int
    amountOfNodes_limit: int
    DeploymentsWithAntiaffinity_length: int
    target_DeploymentsWithAntiaffinity_length: int
    amountOfPodsWithAntiaffinity: int
    target_amountOfPodsWithAntiaffinity: int


    is_node_disruption_searchable: bool
    is_daemonset_disrupted: bool
    is_daemonset_distuption_searchable: bool
    goal_achieved: bool
    block_node_outage_in_progress: bool
    block_policy_calculated : bool
    nodeSelectorsEnabled: bool
    deploymentsWithAntiaffinityBalanced: bool
    maxNumberOfPodsOnSameNodeForDeployment: int
    DeploymentsWithAntiaffinity: Set["mdeployment.Deployment"]
    DeploymentsWithAntiaffinity_length: int
    NodesDrained: Set["mnode.Node"]
    NodesDrained_length: int
    target_NodesDrained_length: int
    target_NodesDrained_length_reached: bool

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_service_disrupted = False
        self.is_deployment_disrupted = False
        self.is_daemonset_disrupted = False
        self.is_node_disrupted = False
        self.goal_achieved = False
        self.is_deployment_distuption_searchable = True
        self.is_service_disruption_searchable = True
        self.is_node_disruption_searchable = True
        self.is_daemonset_distuption_searchable = True
        self.amountOfNodesDisrupted = 0
        self.block_node_outage_in_progress = False
        self.block_policy_calculated = False
        self.antiaffinity_prefered_policy_met = False
        self.nodeSelectorsEnabled = False
        self.amountOfNodes = 0
        self.amountOfNodes_limit = 0
        self.DeploymentsWithAntiaffinity_length = 0
        self.target_DeploymentsWithAntiaffinity_length = 1
        self.amountOfPodsWithAntiaffinity = 0
        self.target_amountOfPodsWithAntiaffinity = 5
        self.deploymentsWithAntiaffinityBalanced = False
        self.maxNumberOfPodsOnSameNodeForDeployment = 10
        self.NodesDrained_length = 0
        self.target_NodesDrained_length = 0
        self.target_NodesDrained_length_reached = False
        
    def __str__(self): return str(self._get_value())


import random
from poodle import Object
from kalc.model.system.base import ModularKind
from kalc.model.system.primitives import TypePolicy
from kalc.misc.const import *
from kalc.misc.util import convertPriorityValue


class PriorityClass(ModularKind):
    metadata_name: str

    priority: int
    preemptionPolicy: TypePolicy

    def __init__(self, *args, **kwargs):
        super().__init__( *args, **kwargs)
        self.metadata_name = "modelPriorityClass"+str(random.randint(100000000, 999999999))
        # self.metadata_name = "model-default-name"
        self.preemptionPolicy = POLICY["PreemptLowerPriority"]
        self.priority = 0

    @property
    def value(self):
        pass
    @value.setter 
    def value(self, value):
        norm_pri = convertPriorityValue(value)
        if norm_pri > 1000: norm_pri = 1000
        self.priority = norm_pri
        
    def __str__(self): return str(self._get_value())

zeroPriorityClass = PriorityClass("ZERO")
zeroPriorityClass.metadata_name = "Normal-zero"



from poodle import planned
from kalc.misc.const import *
import kalc.model.kinds.Service as mservice
from kalc.model.system.base import ModularKind
import kalc.model.system.Scheduler as mscheduler
import kalc.model.kinds.Node as mnode
from kalc.model.kinds.PriorityClass import PriorityClass, zeroPriorityClass
from kalc.model.system.Controller import Controller
from kalc.model.system.primitives import Label

from kalc.model.system.base import HasLimitsRequests, HasLabel
from kalc.misc.const import *
from kalc.misc.util import cpuConvertToAbstractProblem, memConvertToAbstractProblem, getint, POODLE_MAXLIN
# import kalc.cli as cli
import sys
import random
from typing import Set


class Pod(ModularKind, HasLabel, HasLimitsRequests):
    # k8s attributes
    metadata_ownerReferences__name: str
    spec_priorityClassName: str
    spec_nodeSelector: Set[Label]
    metadata_name: str

    metadata_namespace: str
    # internal model attributes
    ownerReferences: Controller
    # TARGET_SERVICE_NULL = mservice.Service.SERVICE_NULL
    # targetService: "mservice.Service"
    atNode: "mnode.Node"
    toNode: "mnode.Node"
    realInitialMemConsumption: int
    realInitialCpuConsumption: int
    currentRealCpuConsumption: int
    currentRealMemConsumption: int
    spec_nodeName: str
    priorityClass: PriorityClass
    status: StatusPod
    isNull: bool
    # amountOfActiveRequests: int # For requests
    hasDeployment: bool
    hasService: bool
    hasDaemonset: bool
    not_on_same_node: Set["Pod"]
    searchable: bool
    

    nodesSelectorSet: bool
    nodeSelectorList: Set["mnode.Node"]
    affinity_set: bool
    antiaffinity_set: bool
    antiaffinity_preferred_set: bool
    
    affinity_met: bool   
    antiaffinity_met: bool
    antiaffinity_preferred_met: bool

    podsMatchedByAffinity: Set["Pod"]
    podsMatchedByAffinity_length: int
    podsMatchedByAntiaffinity: Set["Pod"]
    podsMatchedByAntiaffinity_length: int
    nodesThatCantAllocateThisPod: Set["Node"]
    nodesThatCantAllocateThisPod_length: int
    nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod: Set["Node"]
    calc_nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod: Set["Node"]
    nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod_length: int
    podsMatchedByAntiaffinityPrefered: Set["Pod"]
    podsMatchedByAntiaffinityPrefered_length: int
    calc_affinity_pods_list: Set["Pod"]
    calc_affinity_pods_list_length: int
    calc_antiaffinity_pods_list: Set["Pod"]
    calc_antiaffinity_pods_list_length: int
    calc_cantmatch_antiaffinity: bool
    calc_cantmatch_affinity: bool

    calc_antiaffinity_preferred_pods_list: Set["Pod"]
    calc_antiaffinity_preferred_pods_list_length: int
    antiaffinity_pods_list: Set["Pod"]
    antiaffinity_preferred_pods_list: Set["Pod"]
    
    target_number_of_antiaffinity_pods: int
    target_number_of_antiaffinity_preferred_pods: int

    antiaffinity_labels: Set["Label"]
    antiaffinity_labels_length: int
    antiaffinity_preferred_labels: Set["Label"]
    antiaffinity_preferred_labels_length: int
    affinity_labels: Set["Label"]
    affinity_labels_length: int
    is_checked_as_source_for_labels: bool
    is_checked_as_target_for_labels: bool


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.priority = 0
        self.spec_priorityClassName = "KUBECTL-VAL-NONE"
        self.priorityClass = zeroPriorityClass
        # self.targetService = mservice.Service.SERVICE_NULL
        self.toNode = mnode.Node.NODE_NULL
        self.atNode = mnode.Node.NODE_NULL
        self.cpuRequest = -1
        self.memRequest = -1
        self.status = STATUS_POD["Pending"]
        self.isNull = False
        self.searchable = False
        self.realInitialMemConsumption = 0
        self.realInitialCpuConsumption = 0
        self.currentFormalMemConsumption = 0
        self.currentFormalCpuConsumption = 0
        # self.amountOfActiveRequests = 0 # For Requests
        self.hasService = False
        self.hasDaemonset = False
        self.hasDeployment = False
        self.metadata_name = "modelPod"+str(random.randint(100000000, 999999999))
        # self.metadata_name = "model-default-name"
        self.nodeSelectorSet = False
        self.antiaffinity_set = False
        self.calc_antiaffinity_pods_list_length = 0
        self.calc_antiaffinity_preferred_pods_list_length = 0
        self.target_number_of_antiaffinity_pods = 0
        self.antiaffinity_met = False
        self.podsToBeMatchedByAntiaffinity_length = 0
        self.podsToBeMatchedByAntiaffinity_Preferred_length = 0
        self.podsMatchedByAffinity_length = 0
        self.podsMatchedByAntiaffinity_length = 0
        self.podsMatchedByAntiaffinityPrefered_length = 0
        self.is_checked_as_source_for_labels = False
        self.is_checked_as_target_for_labels = False
        self.nodesThatCantAllocateThisPod_length = 0 
        self.nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod_length = 0
        self.calc_cantmatch_antiaffinity = False
        self.calc_cantmatch_affinity = False




    def set_priority(self, object_space, controller):
        if str(controller.spec_template_spec_priorityClassName) != "Normal-zero":
            try:
                self.priorityClass = next(filter(lambda x: isinstance(x, PriorityClass) and \
                            str(x.metadata_name) == str(controller.spec_template_spec_priorityClassName), object_space))
            except StopIteration:
                logger.warning("Could not reference priority class %s %s" % (str(controller.spec_template_spec_priorityClassName), str(self.metadata_name)))

    def hook_after_load(self, object_space, _ignore_orphan=False):
        pods = list(filter(lambda x: isinstance(x, Pod), object_space))
        nodes = list(filter(lambda x: isinstance(x, mnode.Node) and self.spec_nodeName == x.metadata_name, object_space))
        found = False
        for node in nodes:
            if str(node.metadata_name) == str(self.spec_nodeName):
                self.atNode = node
                node.allocatedPodList.add(self)
                node.allocatedPodList_length += 1
                node.directedPodList.add(self)
                node.directedPodList_length += 1
                node.amountOfActivePods += 1
                assert getint(node.amountOfActivePods) < POODLE_MAXLIN, "Pods amount exceeded max %s > %s" % (getint(node.amountOfActivePods), POODLE_MAXLIN) 
                if self.cpuRequest > 0:
                    node.currentFormalCpuConsumption += self.cpuRequest
                    assert getint(node.currentFormalCpuConsumption) < POODLE_MAXLIN, "CPU request exceeded max: %s" % getint(node.currentFormalCpuConsumption)
                if self.memRequest > 0:
                    node.currentFormalMemConsumption += self.memRequest
                    assert getint(node.currentFormalMemConsumption) < POODLE_MAXLIN, "MEM request exceeded max: %s" % getint(node.currentFormalMemConsumption)
                found = True
            if not self.nodeSelectorSet: self.nodeSelectorList.add(node)
        if not found and self.toNode == mnode.Node.NODE_NULL and not _ignore_orphan:
            logger.warning("Orphan Pod loaded %s" % str(self.metadata_name))
        
        # link service <> pod
        
        services = filter(lambda x: isinstance(x, mservice.Service), object_space)
        for service in services:
            if len(service.spec_selector._get_value()) and \
                    set(service.spec_selector._get_value())\
                        .issubset(set(self.metadata_labels._get_value())):
                # print("ASSOCIATE SERVICE", str(self.metadata_name), str(service.metadata_name))
                service.podList.add(self)
                target_service_of_pod_localVar = service
                self.hasService = True
                if list(service.metadata_labels._get_value()):
                    if self.status._property_value == STATUS_POD["Running"]:
                        service.amountOfActivePods += 1
                        service.status = STATUS_SERV["Started"]
                        assert getint(service.amountOfActivePods) < POODLE_MAXLIN, "Service pods overflow"
                else:
                    pass
                    # cli.click.echo("    - WRN: no label for service %s" % str(service.metadata_name))
        if self.status._property_value == STATUS_POD["Pending"]:
            scheduler = next(filter(lambda x: isinstance(x, mscheduler.Scheduler), object_space))
            scheduler.queueLength += 1
            assert getint(scheduler.queueLength) < POODLE_MAXLIN, "Queue length overflow {0} < {1}".format(getint(scheduler.queueLength), POODLE_MAXLIN)
            scheduler.podQueue.add(self)
            scheduler.status = STATUS_SCHED["Changed"]
            target_service_of_pod_localVar.amountOfPodsInQueue += 1

        nodes = list(filter(lambda x: isinstance(x, mnode.Node), object_space))
        for node in nodes:
            if list(node.metadata_labels) and \
                    set(self.spec_nodeSelector)\
                        .issubset(set(node.metadata_labels)):
                        self.nodeSelectorList.add(node)
                        self.nodeSelectorSet = True

        if str(self.spec_priorityClassName) != "KUBECTL-VAL-NONE":
            try:
                self.priorityClass = next(filter(lambda x: isinstance(x, PriorityClass)\
                    and str(x.metadata_name) == str(self.spec_priorityClassName), \
                        object_space))
            except StopIteration:
                raise Exception("Could not find priorityClass %s, maybe you \
                        did not dump PriorityClass?" % str(self.spec_priorityClassName))
    def hook_after_create(self, object_space):
        nodes = list(filter(lambda x: isinstance(x, mnode.Node), object_space))
        for node in nodes:
            if list(node.metadata_labels) and \
                    set(self.spec_nodeSelector)\
                        .issubset(set(node.metadata_labels)):
                        self.nodeSelectorList.add(node)
                        self.nodeSelectorSet = True
        for node in nodes:
            if not self.nodeSelectorSet: self.nodeSelectorList.add(node)
    @property
    def spec_containers__resources_requests_cpu(self):
        pass
    @spec_containers__resources_requests_cpu.setter
    def spec_containers__resources_requests_cpu(self, res):
        if self.cpuRequest == -1: self.cpuRequest = 0
        self.cpuRequest += cpuConvertToAbstractProblem(res)

    @property
    def spec_containers__resources_requests_memory(self):
        pass
    @spec_containers__resources_requests_memory.setter
    def spec_containers__resources_requests_memory(self, res):
        if self.memRequest == -1: self.memRequest = 0
        self.memRequest += memConvertToAbstractProblem(res)

    @property
    def status_phase(self):
        pass
    @status_phase.setter
    def status_phase(self, res):
        self.status = STATUS_POD[res]

    # we just ignore priority for now
    # @property
    # def spec_priority(self):
    #     pass
    # @spec_priority.setter
    # def spec_priority(self, value):
    #     if value > 1000: value = 1000
    #     self.priority = value
    def connect_pod_service_labels(self,
            pod: "Pod",
            service: "mservice.Service",
            label: Label):
        # TODO: full selector support
        # TODO: only if pod is running, service is started
        assert label in pod.metadata_labels
        assert label in service.spec_selector
        assert pod.status == STATUS_POD["Running"]
        # service.podList.add(pod)
        service.amountOfActivePods += 1
        service.status = STATUS_SERV["Started"]


    def __str__(self): return str(self.metadata_name)




from poodle import Object
import operator as op
from functools import reduce
    # Python3 program to calculate nPr  
import math 
class Combinations(Object):
    number: int
    combinations: int
    def __init__(self, *args, **kwargs):
        super().__init__( *args, **kwargs)
        self.number = 0
        self.combinations = 0
    
    def __str__(self): return str(self.combinations)

def ncr(n, r):
    r = min(r, n-r)
    numer = reduce(op.mul, range(n, n-r, -1), 1)
    denom = reduce(op.mul, range(1, r+1), 1)
    return numer / denom

combinations_list = []
for i in range(10):
    combinations0 = Combinations(i)
    combinations0.number = i
    combinations0.combinations = int(ncr(i,2))
    combinations_list.append(combinations0)

class Permutations(Object):
    number: int
    permutations: int
    def __init__(self, *args, **kwargs):
        super().__init__( *args, **kwargs)
        self.number = 0
        self.permutations = 0
    
    def __str__(self): return str(self.permutations)


def fact(n):  
    if (n <= 1): 
        return 1
          
    return n * fact(n - 1)  
  
def nPr(n, r):  
    
    return math.floor(fact(n) /
                fact(n - r))  
      
permutation_list = []
for i in range(10):
    permutation0 = Permutations(i)
    permutation0.number = i
    permutation0.permutations = int(nPr(i,2))
    permutation_list.append(permutation0)


from typing import Set
from kalc.model.system.base import HasLabel
from kalc.model.system.primitives import Label

class Controller(HasLabel):
    "Kubernetes controller base class"
    spec_selector_matchLabels: Set[Label]


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kalc',
 'kalc.misc',
 'kalc.model',
 'kalc.model.kinds',
 'kalc.model.system',
 'kalc.model.system..ipynb_checkpoints',
 'kalc.policies']

package_data = \
{'': ['*']}

install_requires = \
['Click>=7.0,<8.0',
 'PyYAML>=5.1.2,<6.0.0',
 'Pygments>=2.5,<3.0',
 'jsonpatch>=1.24,<2.0',
 'logzero>=1.5.0,<2.0.0',
 'pandas>=0.25.3,<0.26.0',
 'poodle>=0.1.1',
 'setuptools>=45.0.0,<46.0.0',
 'yaspin>=0.15.0,<0.16.0']

entry_points = \
{'console_scripts': ['kalc-optimize = kalc.misc.cli_optimize:optimize_cluster']}

setup_kwargs = {
    'name': 'kalc',
    'version': '0.1.1',
    'description': 'Kalc - the Kubernetes Calculator core',
    'long_description': '# kalc, the Kubernetes calculator\n\n[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://opensource.org/licenses/Apache-2.0) [![PyPI version](https://badge.fury.io/py/kubectl-val.svg)](https://badge.fury.io/py/kubectl-val) [![Build Status](https://travis-ci.org/criticalhop/kubectl-val.svg?branch=master)](https://travis-ci.org/criticalhop/kubectl-val)\n\n# Overview\n\n`kalc` is a Kubernetes operations automation system built on pure [AI planning](https://github.com/criticalhop/poodle).\n\nCurrently, `kalc` provides a single command, `kalc-optimize` to do automatic cluster rebalancing and optimization.\n\n`kalc` can also be used with [Jupyter Lab](https://jupyter.org/) in interactive mode.\n\n# Quick Start\n\n## Requirements\n\n- Linux x86_64\n- Python 3.7+\n- 6+ GB RAM, decent CPU\n- Up to 20GB of disk space in `/tmp` for generated models\n- `kubectl` installed and connected to cluster\n\n## Installation\n\n    pip install kalc\n\n## Basic usage\n\n    $ kalc-optimize\n\n`kalc-optimize` will generate `bash` scripts containing `kubectl` commands to get to more optimal states. Have a look at those scripts and execute any one of them, then stop and re-run `kalc-optimize`.\n\n## Autopilot\n\n`kalc` can optimize your cluster in background, gradually increasing reliability and cost.\n\n# Architecture\n\n- `kalc-optimize` will download current cluster state by executing `kubectl get all` and will start generating `bash` scripts into current folder\n- Each generated `bash` script contains a sequence of `kubectl` commands to get the cluster in a more optimal state: better balanced nodes for availability and OOM/eviction resilience and a more compact packing\n- As `kalc` continues to compute, it will emit more optimal states and bigger bash scripts with kubectl commands\n\n`kalc` aims to take into account current policies, anti-affility, SLO levels and best practices from successful production Kubernetes clusters.\n\n# Project Status\n\n`kalc` is a developer preview and currently supports a subset of Kubernetes resources and behaviour model.\n\nWe invite you to follow [@criticalhop](https://twitter.com/criticalhop) on [Twitter](https://twitter.com/criticalhop) and to chat with the team at `#kalc` on [freenode](https://freenode.net/). If you have any questions or suggestions - feel free to open a [github issue](https://github.com/criticalhop/kalc/issues) or contact andrew@criticalhop.com directly.\n\nFor enterprise enquiries, use the form on our website: [kalc.io](https://kalc.io) or write us an email at info@kalc.io\n',
    'author': 'CriticalHop Team',
    'author_email': 'info@criticalhop.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



from poodle import Object
from kalc.model.system.primitives import StatusServ, StatusNode, StatusPod, StatusLim, StatusSched, StatusReq, TypePolicy, TypeServ, StatusDeployment, StatusDaemonSet


STATUS_POD = {
    "Running" : StatusPod("Running"),
    "Pending" : StatusPod("Pending"),
    "Killing" : StatusPod("Killing"),
    "Failed" : StatusPod("Failed"),
    "Succeeded" : StatusPod("Succeeded"),
    "Outaged" : StatusPod("Outaged")
}

STATUS_LIM = {
    "Limit Met" : StatusLim("Limit is met"),
    "Limit Exceeded" : StatusLim("Limit is exceded")
}

STATUS_NODE = {
    "Active" : StatusNode("Active"),
    "Inactive" : StatusNode("Inactive"),
    "Killing" : StatusNode("Killing"),
    "New" : StatusNode("New")
    }

STATUS_SCHED = {
    "Changed" : StatusSched("Changed"),
    "Clean" :  StatusSched("Clean")
}

STATUS_SERV = {
    "Interrupted" : StatusServ("Interrupted"),
    "Pending" : StatusServ("Pending"),
    "Started" : StatusServ("Started")
}
POLICY = {
    "PreemptLowerPriority" : TypePolicy("PreemptLowerPriority"),
    "Never" : TypePolicy("Never")
}
STATUS_DEPLOYMENT = {
    "Interrupted" : StatusDeployment("Interrupted"),
    "Pending" : StatusDeployment("Pending"),
    "Started" : StatusDeployment("Started")
}

STATUS_DAEMONSET_STARTED = StatusDaemonSet("Started")
STATUS_DAEMONSET_PENDING = StatusDaemonSet("Pending")
STATUS_DAEMONSET_INTERRUPTED = StatusDaemonSet("Interrupted")



from collections.abc import Mapping, Set, Sequence 
from kalc.misc.object_factory import labelFactory 
import string

CPU_DIVISOR = 65
MEM_DIVISOR = 200

from poodle.arithmetic import logSparseIntegerFactory
for n in range(1,10000):
    try:
        i = logSparseIntegerFactory.numbers[n]
    except KeyError:
        break
POODLE_MAXLIN = n - 1

PRIO_MAPPING = {i: i for i in range(POODLE_MAXLIN)}

try:
    from six import string_types, iteritems
except ImportError:
    string_types = (str, ) if str is bytes else (str, bytes)
    iteritems = lambda mapping: getattr(mapping, 'iteritems', mapping.items)()

def objwalk(obj, path=(), memo=None):
    if memo is None:
        memo = set()
    if isinstance(obj, Mapping):
        if id(obj) not in memo:
            memo.add(id(obj)) 
            for key, value in iteritems(obj):
                for child in objwalk(value, path + (key,), memo):
                    yield child
    elif isinstance(obj, (Sequence, Set)) and not isinstance(obj, string_types):
        if id(obj) not in memo:
            memo.add(id(obj))
            for index, value in enumerate(obj):
                for child in objwalk(value, path + (index,), memo):
                    yield child
    else:
        yield path, obj

def find_property(obj, p):
    path = p[0]
    val = p[1]
    if len(path) == 1:
        if hasattr(obj, path[0]):
            return path[0], val
    
    for i in range(len(path), 1, -1):
        try_path = path[:i]
        spath = "_".join([x if isinstance(x, str) else "" for x in try_path])
        # print("Trying ", spath)
        if hasattr(obj, spath) and i==len(path):
            # print("FOUND1")
            return spath, val
        elif hasattr(obj, spath) and i==len(path)-1:
            # print("FOUND2")
            return spath, {path[-1]: val}
    return None, None

def k8s_to_domain_object(obj):
    try_int = False
    try:
        int(obj)
        try_int = True
    except:
        pass
    if isinstance(obj, int):
        return obj
    elif isinstance(obj, dict) and len(obj) == 1:
        k,v=list(obj.items())[0]
        return labelFactory.get(k,v)
    elif isinstance(obj, str) and obj[0] in string.digits+"-" and not obj[-1] in string.digits:
        # pass on, probably someone will take care
        return obj
    elif isinstance(obj, str) and try_int:
        return int(obj)
    elif isinstance(obj, str) and not obj[0] in string.digits+"-":
        return obj
    else:
        return obj.__str__()
        # raise ValueError("Value type not suported: %s" % repr(obj))

def cpuConvertToNorm(cpuParot):
    cpu = 0
    if isinstance(cpuParot, int):
        cpu = cpuParot*1000
    else:
        if cpuParot[len(cpuParot)-1] == 'm':
            cpu = int(cpuParot[:-1])
        else:
            cpu = int(cpuParot)*1000
    return int(cpu)

def memConvertToNorm(mem):
    ret = 0
    if mem[len(mem)-2:] == 'Gi':
        ret = int(mem[:-2])*1000
    elif mem[len(mem)-2:] == 'Mi':
        ret = int(mem[:-2])
    elif mem[len(mem)-2:] == 'Ki':
        ret = int(int(mem[:-2])/1000)
    else:
        ret = int(int(mem)/1000000)
    return int(ret)

def cpuConvertToAbstractProblem(cpuParot):
    #log.debug("cpuParot", cpuParot)
    cpu = 0
    if isinstance(cpuParot, int):
        cpu = cpuParot*1000
    else:
        if cpuParot[len(cpuParot)-1] == 'm':
            cpu = int(cpuParot[:-1])
        else:
            cpu = int(cpuParot)*1000
    # log.debug("cpuParot ", cpuParot, " ret ", cpuAdd)
    cpu = int(cpu / CPU_DIVISOR)
    if cpu == 0:
        cpu = 1
    assert cpu >= 0
    assert cpu <= POODLE_MAXLIN, "Number exceeds Poodle power"
    return int(cpu)

def memConvertToAbstractProblem(mem):
    ret = 0
    if mem[len(mem)-2:] == 'Gi':
        ret = int(mem[:-2])*1000
    elif mem[len(mem)-2:] == 'Mi':
        ret = int(mem[:-2])
    elif mem[len(mem)-2:] == 'Ki':
        ret = int(int(mem[:-2])/1000)
    else:
        ret = int(int(mem)/1000000)
    ret = int(ret / MEM_DIVISOR)
    if ret == 0:
        ret = 1
    assert ret >= 0
    assert ret <= POODLE_MAXLIN, "Number exceeds Poodle power"
    return int(ret)

#object deduplicator by metadata_name
def objDeduplicatorByName(objList):
    dedupList = []
    nameList = []
    counter = 0
    for obj in objList:
        if not(obj.metadata_name._get_value() in nameList):
            dedupList.append(obj)
            nameList.append(obj.metadata_name._get_value())
        counter +=1
    return dedupList

#solve bug in poodle by this
def objRemoveByName(objList, metadata_name):
    br = True
    while br:
        br = False
        for obj in objList:
            if obj.metadata_name._get_value() == metadata_name :
                objList.remove(obj)
                br = True
                break

def poodle_bug_dedup(podList):
    lpods = podList._get_value()
    dd_lpods = []
    for podOb in lpods:
        found = False
        for p in dd_lpods:
            if str(p.metadata_name) == str(podOb.metadata_name):
                found = True
                break
        if found: continue
        dd_lpods.append(podOb)
    return dd_lpods

def getint(poob):
    return int(poob._get_value())

def split_yamldumps(s: str):
    spl = s.split("\nkind: List\n")
    return [x for x in spl if len(x.split("\n")) > 4]

def convertPriorityValue(v: int):
    assert isinstance(v, int)
    if v == 0: return 0
    global PRIO_MAPPING
    return PRIO_MAPPING[v]



from typing import Set
from kalc.model.kinds.Pod import Pod
from kalc.model.kinds.Node import Node
from kalc.model.kinds.Service import Service
from kalc.model.system.Scheduler import Scheduler
from kalc.model.system.globals import GlobalVar
from kalc.misc.const import *
from poodle import planned

from kalc.policy import policy_engine, BasePolicy

class PreferredSelfAntiAffinityPolicy(BasePolicy):
    TYPE = "property"
    KIND = Service

    def register(self):
        Service.register_property(name="antiaffinity", type=bool, default=False)
        Service.register_property(name="antiaffinity_prefered_policy_met", type=bool, default=False)
        Service.register_property(name="targetAmountOfPodsOnDifferentNodes", type=int, default=-1)
        Pod.register_property(name="not_on_same_node", type=Set[Pod], default=None)

    def set(self, val: bool):
        assert isinstance(val, bool), "Can only accept True or False"

        # TODO 1: do we have anti affinity already?
           # 1.1: anti affinity policy satisfied already
           # 1.2: anti affinity is implemented in settings
        # TODO 2: add anti affinity if it is not there
        # TODO 3: check that it will not break cluster
        # TODO 4: find solution to fix cluster if #3 fails

        if val:
            # enable
            pods_count = len(list(self.target_object.podList))

            assert pods_count <= 5, "We currently support up to 5 pods"

            def hypothesis_existing_match_group():
                # TODO: 
                pass

            def hypothesis_new_match_group():
                # TODO: hypotheses can not work in parallel this way: will modify main object

                # TODO: create a new group before start
                self.target_object.antiaffinity = True
                self.target_object.targetAmountOfPodsOnDifferentNodes = pods_count
                self.register_goal(self.target_object.antiaffinity_prefered_policy_met, "==", True)
            
            self.register_hypothesis("All pods required done", hypothesis_1)
        else:
            # disable
            pass

    @planned
    def add_pods_to_antiaffinity_group(self, 
            pod: Pod,
            pod_other: Pod):
        pod.anti_affinite_with.add(pod_other)
        # TODO we need a restart after this to check that it didn't break things

    @planned(cost=1)
    def mark_antiaffinity_prefered_policy_enabled(self,
        service: Service,
        globalVar: GlobalVar):
        assert service.amountOfPodsOnDifferentNodes == service.targetAmountOfPodsOnDifferentNodes
        assert service.antiaffinity == True
        service.antiaffinity_prefered_policy_met = True

    @planned(cost=1)
    def mark_antiaffinity_prefered_policy_met(self,
        service: Service,
        globalVar: GlobalVar):
        assert service.amountOfPodsOnDifferentNodes == service.targetAmountOfPodsOnDifferentNodes
        assert service.antiaffinity == True
        service.antiaffinity_prefered_policy_met = True

    @planned(cost=1)
    def mark_2_pods_of_service_as_not_at_same_node(self,
            pod1: Pod,
            pod2: Pod,
            node_of_pod1: Node,
            node_of_pod2: Node,
            service: Service,
            scheduler: Scheduler):
        assert node_of_pod2 == pod2.atNode
        assert pod1.atNode in node_of_pod2.different_than
        assert pod1 in service.podList
        assert pod2 in service.podList
        assert scheduler.status == STATUS_SCHED["Clean"]
        pod1.not_on_same_node.add(pod2)
        service.amountOfPodsOnDifferentNodes = 2

    @planned(cost=1)
    def mark_3_pods_of_service_as_not_at_same_node(self,
            pod1: Pod,
            pod2: Pod,
            pod3: Pod,
            node_of_pod1: Node,
            node_of_pod2: Node,
            node_of_pod3: Node,
            service: Service,
            scheduler: Scheduler):
        
        assert node_of_pod1 == pod1.atNode
        assert node_of_pod2 == pod2.atNode
        assert node_of_pod3 == pod3.atNode
        assert node_of_pod1 in node_of_pod2.different_than
        assert node_of_pod1 in node_of_pod3.different_than
        assert node_of_pod2 in node_of_pod3.different_than
        assert pod1 in service.podList
        assert pod2 in service.podList
        assert pod3 in service.podList
        assert scheduler.status == STATUS_SCHED["Clean"]
        service.amountOfPodsOnDifferentNodes = 3

    @planned(cost=1)
    def mark_4_pods_of_service_as_not_at_same_node(self,
            pod1: Pod,
            pod2: Pod,
            pod3: Pod,
            pod4: Pod,
            node_of_pod1: Node,
            node_of_pod2: Node,
            node_of_pod3: Node,
            node_of_pod4: Node,
            service: Service,
            scheduler: Scheduler):
        
        assert node_of_pod1 == pod1.atNode
        assert node_of_pod2 == pod2.atNode
        assert node_of_pod3 == pod3.atNode
        assert node_of_pod4 == pod4.atNode
        assert node_of_pod1 in node_of_pod2.different_than
        assert node_of_pod1 in node_of_pod3.different_than
        assert node_of_pod1 in node_of_pod4.different_than
        assert node_of_pod2 in node_of_pod3.different_than
        assert node_of_pod2 in node_of_pod4.different_than
        assert node_of_pod3 in node_of_pod4.different_than
        assert pod1 in service.podList
        assert pod2 in service.podList
        assert pod3 in service.podList
        assert pod4 in service.podList
        assert scheduler.status == STATUS_SCHED["Clean"]
        service.amountOfPodsOnDifferentNodes = 4

    @planned(cost=1)
    def mark_5_pods_of_service_as_not_at_same_node(self,
            pod1: Pod,
            pod2: Pod,
            pod3: Pod,
            pod4: Pod,
            pod5: Pod,
            node_of_pod1: Node,
            node_of_pod2: Node,
            node_of_pod3: Node,
            node_of_pod4: Node,
            node_of_pod5: Node,
            service: Service,
            scheduler: Scheduler):
        
        assert node_of_pod1 == pod1.atNode
        assert node_of_pod2 == pod2.atNode
        assert node_of_pod3 == pod3.atNode
        assert node_of_pod4 == pod4.atNode
        assert node_of_pod5 == pod5.atNode
        assert node_of_pod1 in node_of_pod2.different_than
        assert node_of_pod1 in node_of_pod3.different_than
        assert node_of_pod1 in node_of_pod4.different_than
        assert node_of_pod1 in node_of_pod5.different_than
        assert node_of_pod2 in node_of_pod3.different_than
        assert node_of_pod2 in node_of_pod4.different_than
        assert node_of_pod2 in node_of_pod5.different_than
        assert node_of_pod3 in node_of_pod4.different_than
        assert node_of_pod3 in node_of_pod5.different_than
        assert node_of_pod4 in node_of_pod5.different_than

        assert pod1 in service.podList
        assert pod2 in service.podList
        assert pod3 in service.podList
        assert pod4 in service.podList
        assert pod5 in service.podList
        assert scheduler.status == STATUS_SCHED["Clean"]
        service.amountOfPodsOnDifferentNodes = 5

policy_engine.register(Service, "self_antiaffinity", PreferredSelfAntiAffinityPolicy)




"""Implementing full model from all effects"""
from poodle import Object
import kalc.model.kinds
import pkgutil, inspect
import kalc.misc.problem

kinds_collection = {}

for importer, modname, ispkg in pkgutil.iter_modules(
        path=kalc.model.kinds.__path__,
        prefix=kalc.model.kinds.__name__+'.'):
    module = __import__(modname, fromlist="dummy")
    globals()[modname.split(".")[-1]] = module
    for n in dir(module):
        c = getattr(module, n)
        if inspect.isclass(c) and issubclass(c, Object) and not c is Object:
            kinds_collection[c.__name__] = c

Node.Node.NODE_NULL = Node.Node("NULL")
Node.Node.NODE_NULL.isNull = True
Node.Node.NODE_NULL.metadata_name = "Null-Node"
Node.Node.NODE_NULL.searchable = False

Pod.Pod.POD_NULL = Pod.Pod("NULL")
Pod.Pod.POD_NULL.isNull = True

    